<?php

require 'Flickr/API.php';

/* 
 * Flickr captcha class
 * 
 * Provide the user with a number of images, one of them contains a 
 * tag (normally a noun) that they must identify. The others do not 
 * contain that tag (and contain at least one of the other nouns
 * you have specified
 *
 * filename: flickr.captcha.class.php
 * @author (Copyright) Joel Heenan 13/12/2008
 *
 * This file is part of Flickr-CAPTCHA
 * 
 * Flickr-CAPTCHA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Flickr-CAPTCHA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flickr-CAPTCHA.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ChangeLog
 * 
 * v0.1 - Joel Heenan - 24/12/2008 - Initial release
 */


class flickr_captcha {

    /* @shortdesc Flickr API key
     * See http://www.flickr.com/services/api/misc.api_keys.html
     * @type string
     * @public
     */
    var $apikey = 'USE_YOUR_KEY_HERE';

    /* @shortdesc Tags to select images using. Try to keep them to nouns that don't have
     * too many meanings and don't overlap. beach and sea might cause confusion
     * @type string
     * @public
     */
    var $tags = array('baby','beach','car','cat','dog','flower','girl','lake','tree','wedding');

    /* @shortdesc Number of choices to select from
     * @type string
     * @public
     */
    var $numChoices = 10;

    /* @shortdesc Display errors. Should we display errors to the user? 
     * @type bool
     * @ public
     */
    var $displayErrors = True;

    /* @shortdesc The session key we will use to set the values
     * @type string
     * @public
     */
    var $sessionKey = 'flickr_captcha';

    /* @shortdesc Maximum amount of tries before we fail them
     * @type int
     * @public
     */
    var $maxtry = 3;

    /* @shortdesc the ratio of correct answers we require to consider this a good response
     * @type int
     * @public
     */
    var $acceptedRatio = 0.7;

    /* @shortdesc license ID's that are acceptable. Since we attribute authors and are 
     * non-commercial and non-derivative we can use all license types except for the 
     * default which is all rights restricted
     * @type string
     * @public
     */
    var $licenseIDs = '1,2,3,4,5,6,7';

    /* @shortdesc disclaimer text, required legally by flickr
     * @type string
     * @public
     */
    var $disclaimer = "
This product uses the Flickr API but is not endorsed or certified by Flickr. 
Find the latest version of Flickr-CAPTCHA at <a href='http://planetjoel.com/viewarticle/630/Flickr+Captcha+Using+PHP'>planetjoel.com</a>. 
This product is licensed under the <a href='http://www.gnu.org/licenses/gpl-3.0.txt'>GPLv3</a>.
    ";

    /* PRIVATE */

    /* CONSTANTS */
    var $PASSED          = 1;
    var $FAILED          = 2;
    var $MAXTRY_EXCEEDED = 3;
    var $NO_POST         = 4;

    /* @shortdesc Images the user got correct
     * @type array
     * @public
     */
    var $correct = array();

    /* @shortdesc Images the user got wrong
     * @type array
     * @public
     */
    var $wrong = array();

    /* @shortdesc Images teh user missed
     * @type array
     * @public
     */
    var $missed = array();

    /* @shortdesc API
     * @type Flickr/API object
     * @private
     */
    var $api;

    /* @shortdesc Max tags. Flickr does not allow more than a certain number of tags
     * per search
     * @type int
     * @private
     */
    var $maxTags = 10;

    /* @shortdesc maximum amount of items per row
     * @type int
     * @private
     */
    var $maxColSize = 5;

    /* @shortdesc debug mode 
     * @type bool
     * @private
     */
    var $debug = False;

    /* @shortdesc cache to store owner details
     * @type array
     * @private
     */
    var $personCache = array();

    /* @shortdesc extract config array
     * @private
     * @type void
     * @return nothing
     */
    function flickr_captcha($config=array(),$secure=TRUE) {
        if($secure) {
            $valid = get_class_vars(get_class($this));
            foreach($config as $k=>$v) {
                if(array_key_exists($k,$valid)) $this->$k = $v;
            }
        } else {
            foreach($config as $k=>$v) $this->$k = $v;
        }

        if(!$_SESSION['fc_tries']) {
            $_SESSION['fc_tries'] = 0;
        }

        $this->_setupAPI();
    }

    /* @shortdesc validates POST and returns the result. Human after all?
     * @public
     * @type integer
     * @return 
     */
    function validate_submit() {
        if(!$_POST['selected']) {
            return $this->NO_POST;
        }
        
        $userSelected = split(",", $_POST['selected']);
        $answers = split(",", $_SESSION[$this->sessionKey]);

        $this->_debug(sprintf("Answers: %s",join(",",$answers)));

        $tries = $_SESSION['fc_tries'];
        $_SESSION['fc_tries']++;
        if($tries > $this->maxtry) {
            return $this->MAXTRY_EXCEEDED;
        }

        foreach($userSelected as $selected) {
            $this->_debug(sprintf("Selected: %s",$selected));
            if(!is_numeric($selected)) {
                $this->_debug(sprintf("NAN: %s",$selected));
                continue;
            }
            if(empty($selected)) {
                $this->_debug(sprintf("Empty: %s",$selected));
                continue;
            }

            if(in_array($selected,$answers)) {
                array_push($this->correct,$selected);
            } else {
                array_push($this->wrong,$selected);
            }
        }

        foreach($answers as $answer) {
            if(!in_array($answer,$userSelected)) {
                array_push($this->missed,$answer);
            }
        }

        $divider = count($this->correct) - count($this->wrong) - count($this->missed);
        if($divider == 0) {
            $ratio = 0;
        } else {
            $ratio = $this->numChoices / $divider;
        }
        $this->_debug("Correct: " . count($this->correct));
        $this->_debug("Wrong: " . count($this->wrong));
        $this->_debug("Missed: " . count($this->missed));
        if( $ratio > $this->acceptedRatio ) {
            $_SESSION['fc_tries'] = 0;
            return $this->PASSED;
        } else {
            return $this->FAILED;
        }
    }

    /* @shortdesc Return HTML showing what images the user missed or got wrong
     * @type void
     * @return HTML
     */
    function display_report() {
        /* TODO: code this bitch */
        $res = "";

        if(count($this->wrong) > 0) {
            $res .= "<p>Did not contain the tag:</p>";
            foreach($this->wrong as $w) {
                $url = $_SESSION['fc_'.$w];
                $res .= sprintf("<img src='%s'>, ",$url);
            }
        }

        if(count($this->missed) > 0) {
            $res .= "<p>You missed these:</p>";
            foreach($this->missed as $m) {
                $url = $_SESSION['fc_'.$m];
                $res .= sprintf("<img src='%s'>, ",$url);
            }
            $res .= "<br/>";
            return $res;
        }

        if((count($this->wrong) === 0) and (count($this->missed) === 0)) {
            $res .= "<p>Perfect!</p>";
        }

        return $res;
    }

    /* @shortdesc display the form
     * @public
     * @type void
     * @arguments insideForm - don't output a new form
     * @return HTML-Output
     */
    function display_form($insideForm = False) {
        $this->_debug("display_form()");
        $numGoodImages = rand(1, floor($this->numChoices - $this->numChoices*0.3));
        $this->_debug(sprintf("numGoodImages %s",$numGoodImages));
        $randkey = array_rand($this->tags,1);
        $keyTag = $this->tags[$randkey];
        $otherTags = $this->_getOtherTags($keyTag);
        $goodImages = $this->getImages(array($keyTag), $numGoodImages);
        $decoyImages = $this->getImages($otherTags, $this->numChoices - $numGoodImages,$keyTag,$goodImages);
        if(!$goodImages or !$decoyImages) {
            return $this->handleError();
        } 
        $this->_debug(sprintf("Got %s good images, %s decoy images",count($goodImages),count($decoyImages)));
        $this->_setSessionKey($goodImages);
        return $this->_outputHTML($goodImages,$decoyImages,$keyTag,$insideForm); 
    }

    /* @shortdesc Print stuff about the error if enabled
     * @public
     * @type void
     * @return HTML-Output
     */
    function handleError() {
        $res = "";
        if(!$this->displayErrors) {
            $res .= "<!-- ";
        }
        if($this->error) {
            $res .= sprintf("Internal Error: %s",$this->error);
        } else {
            $res .= sprintf("Flickr API Error %s: %s",
                            $this->api->getErrorCode(),$this->api->getErrorMessage());
        }
        if(!$this->displayErrors) {
            $res .= "-->";
        }
        return $res;
    }
   
    /* @shortdesc given an array of tags and a number of images
     * return a random selection of them
     * @type array
     * @public
     * @return a selection of images
     */
    function getImages($tags, $num, $excludeTag='', $alreadyChosen=array()) {
        if(!$tags) {
            $this->error = 'No tags passed to getImages';
            return False;
        }
        
        $res = array();

        $tagsStr = join(",",$tags);

        $response = $this->api->callMethod('flickr.photos.search', array(tags=>$tagsStr,license=>$this->licenseIDs));
        if(!$response) {
            print $this->handleError();
            return False;
        }

        $responseImages = $response->children[1]->children;
        while(count($res) < $num) {
            $randkey = array_rand($responseImages,1);
            $image = $responseImages[$randkey];
            if($image->attributes['id']) {
                foreach($alreadyChosen as $chosenImage) {
                    if($chosenImage->id == $image->attributes['id']) {
                        continue;
                    }
                }
                if($excludeTag) {
                    $imageTags = $this->getImageTags($image->attributes['id'],$image->attributes['secret']);
                    if(!$imageTags) {
                        return False;
                    }
                    if(in_array($excludeTag,$imageTags)) {
                        continue;
                    }
                }
                $fi = new FlickrImage($image);
                array_push($res,$fi);
                $this->_debug(sprintf('Setting session %s to %s','fc_'.$fi->id,$fi->url));
                $_SESSION['fc_'.$fi->id] = $fi->url;
                unset($responseImages[$randkey]);
            }
        }

        return $res;
    }

    /* @shortdesc return all tags for an image
     * @type array
     * @public
     * @return array of tags
     */
    function getImageTags($imageId, $secret) {
        if(!$imageId) {            
            $this->error = 'no image id passed to getImageTags';
            return False;
        }

        $response = $this->api->callMethod('flickr.photos.getInfo', 
                                            array(photo_id=>$imageId,secret=>$secret));
        if(!$response) {
            print $this->handleError();
            return False;
        }

        $res = array();

        foreach($response->children[1]->children as $item) {
            if($item->name == 'tags') {
                foreach($item->children as $tag) {
                    if($tag->name == 'tag') {
                        array_push($res,$tag->content);
                    }
                }
            }
        }

        return $res;
    }

    /* @shortdesc given a NSID return a person object. Cached for 
     * the duration of the object
     * @public
     * @type FlickrPerson
     */
    function getPerson($nsid) {
        $this->_debug("Trying to get person " . $nsid);
        if($personCache[$nsid]) {
            return $personCache[$nsid];
        }

        $response = $this->api->callMethod('flickr.people.getInfo',
                                            array(user_id=>$nsid));

        if(!$response) {
            print $this->handleError();
            return False;
        }

        $personCache[$nsid] = new FlickrPerson($response);
        return $personCache[$nsid];
    }

    /* Private deep dark ju-ju */
    
    /* @shortdesc connect to the Flickr API
     * @private
     * @type void
     * @return nothing
     */
    function _setupAPI() {
        $this->api = new Flickr_API(array('api_key' => $this->apikey));
    }

    /* @shortdesc Given a keytag, exclude that from the list
     * then return a random list of a few of the other tags
     * @type array
     * @return a list of random tags not including the keytag
     */
    function _getOtherTags($keyTag) {
        $othertags = $this->tags;

        foreach($othertags as $k=>$v) {
            if($v == $tag) {
                unset($othertags[$k]);
            }
        }
        
        shuffle($othertags);
        return array_slice($othertags,0,$this->maxTags);
    }

    /* @shortdesc Set the session key so that we can later work out whether the person
     * got the results correct
     * @type void
     * @return none
     */
    function _setSessionKey($goodImages) {
        $ids = array();
        foreach($goodImages as $i) {
            array_push($ids,$i->id);
        }
        $v = join(",",$ids);

        $_SESSION[$this->sessionKey] = $v;
        $this->_debug(sprintf("Setting session key to %s", $v));
    }

    /* @shortdesc Given some good images and some bad images, shuffle them 
     * together and display them to the user
     * @private
     * @return HTML to be displayed
     */
    function _outputHTML($goodImages, $badImages,$keyTag,$insideForm) {
        $allImages = array_merge($goodImages, $badImages);
        shuffle($allImages);

        $temp = $this->_determineRowColSize();
        $rowsize = $temp[0];
        $colsize = $temp[1];
        $owner = $this->getPerson($allImages[0]->owner);
       
        $res = "";
        $res .= "<div class='fc_Mainbox'>\n";
        $res .= sprintf("Please select all images you associate with <span class='fc_Keytag'>%s</span>",$keyTag);;
        $res .= "<div class='fc_Windowbox'>";
        $res .= sprintf("<img class='fc_previewImage' src='%s' id='previewImage'>\n",$allImages[0]->url);
        $text = sprintf("%s - <a href='%s'>%s</a>",$allImages[0]->title,$owner->photosurl,$owner->username);
        $res .= sprintf("<div id='previewTitle'>%s</div>\n",$text);
        $res .= "</div>";
        $res .= "<table>\n";

        for($i=0; $i<$colsize; $i++) {
            $res .= "\t<tr>";
            for($j=0; $j<$rowsize; $j++) {
                $image = $allImages[($i*$rowsize)+$j];
                $owner = $this->getPerson($image->owner);
                $showPreview = sprintf("showPreview(\"fc_Picture_%s\",\"%s\",\"%s\",\"%s\",\"%s\")",$image->id,$image->url,$image->title,$owner->username,$owner->photosurl);
                $hidePreview = sprintf("hidePreview(\"fc_Picture_%s\")",$image->id);
                $selectImage = sprintf("selectImage(\"fc_Picture_%s\",\"%s\")",$image->id,$image->id);
                $res .= sprintf("<td><div class='fc_Picture' id='fc_Picture_%s' onclick='%s'>",$image->id,$selectImage);
                $res .= sprintf("<img width=50 src='%s' onMouseOver='%s' onMouseOut='%s'>",
                                $image->url,$showPreview,$hidePreview,$selectImage);
                $res .= "</div></td>\n";
            }
            $res .= "\t<br/></tr>\n";
        }

        $res .= "</table>\n";

        if(!$insideForm) {
            $res .= "<form method='post' name='fc_form'>\n";
        }
        $res .= "<input type='hidden' name='selected' value=''>\n";
        if(!$insideForm) {
            $res .= "<center><input type='submit' value='submit'></center>\n";
            $res .= "</form>\n";
        }
        $res .= sprintf("<span class='fc_disclaimer'>%s</span>",$this->disclaimer);
        $res .= "</div>";

        return $res;
    }

    /* @shortdesc try to find a row size that is balanced and no more than
     * maxRowSize items per row
     * @type array
     * @private
     * @return (row size, col size)
     */
    function _determineRowColSize() {
        $rowsize = 6;
        $divider = 1;
        while($rowsize > 5) {
            $divider += 1;
            $rowsize = floor($this->numChoices / $divider);
        }
        $this->_debug(sprintf("Got a row size of %s col size of %s",$rowsize,$divider));
        return array($rowsize,$divider);
    }

    /* @shortdesc print something but only if we are in debug mode
     * @type void
     * @private
     */
    function _debug($msg) { 
        if($this->debug) {
            echo "d: $msg<br/>\n";
        }
    }
}

/* @shortdesc a class representing a flickr image
 */
class FlickrImage {

    /* @shortdesc see http://www.flickr.com/services/api/misc.urls.html
     * @type string
     * @public
     */
    var $size = "m";
    
    /* @shortdesc given some XML data parse the flickr image details
     */
    function FlickrImage($xmlData) {
        foreach($xmlData->attributes as $k=>$v) $this->$k = $v;
        $this->url = $this->buildUrl();
    }

    function buildUrl() {
        return sprintf("http://farm%s.static.flickr.com/%s/%s_%s_%s.jpg", 
                        $this->farm, $this->server, $this->id, $this->secret, $this->size);
    }
}

/* @shortdesc a class representing a flickr person
 */
class FlickrPerson {
    /* @shortdesc given XML data parse person details
     */
    function FlickrPerson($xmlData) {
        foreach($xmlData->children[1]->attributes as $k=>$v) $this->$k = $v;
        foreach($xmlData->children[1]->children as $field) {
            $name = $field->name;
            if(!empty($name)) {
                $this->$name = $field->content;
            }
        }
    }
}

?>
